import logging
from pathlib import Path

import alembic.command
import alembic.config

from ...configuration import config

log = logging.getLogger(__name__)

HERE = Path(__file__).parent


class AlembicMigration:
    """Glue Duffy (mainly its config) and Alembic DB migrations together."""

    @property
    def config(self):
        if not hasattr(self, "_config"):
            self._config = alembic.config.Config()
            self._config.set_main_option("script_location", str(HERE.absolute()))
            self._config.set_main_option(
                "sqlalchemy.url", config["database"]["sqlalchemy"]["sync_url"]
            )
        return self._config

    def create(self, comment: str, autogenerate: bool):
        alembic.command.revision(config=self.config, message=comment, autogenerate=autogenerate)

        if autogenerate:
            log.warning(
                ">>> Remember to edit the autogenerated migration script! Unedited, it would drop"
                " any support table not registered with your SQLAlchemy metadata."
            )

    def db_version(self):
        alembic.command.current(self.config)

    def upgrade(self, version: str):
        alembic.command.upgrade(self.config, version)

    def downgrade(self, version: str):
        alembic.command.downgrade(self.config, version)


alembic_migration = AlembicMigration()



import datetime as dt

from sqlalchemy import JSON, Boolean, Column, ForeignKey, Index, Integer, Text, UnicodeText, and_
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship

from ...api_models import SessionNodeModel
from .. import Base
from ..types import NodeState
from ..util import CreatableMixin, RetirableMixin
from .session import Session


class Node(Base, CreatableMixin, RetirableMixin):
    index_uniqueness_clause = and_(
        Column("retired_at") == None,  # noqa: E711
        Column("state") != "provisioning",
    )

    __tablename__ = "nodes"
    __table_args__ = (
        Index(
            "active_hostname_index",
            "hostname",
            unique=True,
            sqlite_where=index_uniqueness_clause,
            postgresql_where=index_uniqueness_clause,
        ),
        Index(
            "active_ipaddr_index",
            "ipaddr",
            unique=True,
            sqlite_where=index_uniqueness_clause,
            postgresql_where=index_uniqueness_clause,
        ),
    )

    id = Column(Integer, primary_key=True, nullable=False)
    hostname = Column(Text, nullable=True)
    ipaddr = Column(Text, nullable=True)
    state = Column(
        NodeState.db_type(),
        nullable=False,
        default=NodeState.unused,
        server_default=NodeState.unused.value,
    )
    comment = Column(UnicodeText, nullable=True)

    pool = Column(UnicodeText, nullable=True, index=True)

    reusable = Column(Boolean, nullable=False, default=False, server_default="FALSE")

    # Careful, MutableDict only detects changes to the top level of dict key-values!
    data = Column(
        MutableDict.as_mutable(JSON), nullable=False, default=lambda: {}, server_default="{}"
    )

    def fail(self, detail: str):
        """Set the state of a node to failed with details"""
        self.state = NodeState.failed
        self.data["error"] = {"failed_at": dt.datetime.utcnow().isoformat(), "detail": detail}


class SessionNode(Base):
    __tablename__ = "sessions_nodes"
    session_id = Column(Integer, ForeignKey(Session.id), primary_key=True, nullable=False)
    session = relationship(Session, back_populates="session_nodes")
    node_id = Column(Integer, ForeignKey(Node.id), primary_key=True, nullable=False)
    node = relationship(Node)

    pool = Column(UnicodeText, nullable=False, index=True)

    # Careful, MutableDict only detects changes to the top level of dict key-values!
    data = Column(
        MutableDict.as_mutable(JSON), nullable=False, default=lambda: {}, server_default="{}"
    )

    @property
    def pydantic_view(self) -> SessionNodeModel:
        args = {
            "id": self.node.id,
            "hostname": self.node.hostname,
            "ipaddr": self.node.ipaddr,
            "pool": self.node.pool,
            "reusable": self.node.reusable,
            "data": self.node.data,
        }

        if self.session.active:
            args["state"] = self.node.state

        return SessionNodeModel(**args)



from .util import DeclEnum


class NodeState(str, DeclEnum):
    unused = "unused"
    provisioning = "provisioning"
    ready = "ready"
    contextualizing = "contextualizing"
    deployed = "deployed"
    deprovisioning = "deprovisioning"
    done = "done"
    failed = "failed"



import re
from datetime import timedelta
from functools import lru_cache
from typing import Type

from pydantic.types import _registered


class _TimeDelta(timedelta):
    strip_whitespace = True
    to_lower = True
    needs_sign = False
    allow_negative = True
    allow_dimensionless_seconds = True

    def __init_subclass__(cls):
        sign = "+-" if cls.allow_negative else "+"
        needs_sign = "" if cls.needs_sign else "?"
        cls.timedelta_re = re.compile(
            rf"(?P<sign>[{sign}]){needs_sign}(?!\s*$)"
            + r"(?:\s*(?P<weeks>\d+)w)?"
            + r"(?:\s*(?P<days>\d+)d)?"
            + r"(?:\s*(?P<hours>\d+)h)?"
            + r"(?:\s*(?P<minutes>\d+)m)?"
            + r"(?:\s*(?P<seconds>\d+)s)?"
            + r"(?:\s*(?P<milliseconds>\d+)ms)?"
        )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(examples=["+3h30m"])
        if cls.needs_sign:
            field_schema.update(pattern=r"^[-+](\d+w)?(\d+d)?(\d+h)?(\d+m)?(\d+s)?(\d+ms)?$")
        else:
            field_schema.update(pattern=r"^[-+]?(\d+w)?(\d+d)?(\d+h)?(\d+m)?(\d+s)?(\d+ms)?$")
            field_schema["examples"].append("1h")

        if cls.allow_negative:
            field_schema["examples"].append("-20m25s")

    @classmethod
    @lru_cache
    def validate(cls, v):
        if isinstance(v, timedelta):
            return v

        if cls.allow_dimensionless_seconds:
            if isinstance(v, int):
                return timedelta(seconds=v)
            if not isinstance(v, str):
                raise TypeError(f"input value {v!r} must be a string or an integer")
            if v.isdigit():
                return timedelta(seconds=int(v))
        elif not isinstance(v, str):
            raise TypeError("input value must be a string")
        m = cls.timedelta_re.fullmatch(v.strip().lower())
        if not m:
            raise ValueError("invalid timedelta format")
        sign = -1 if m.group("sign") == "-" else 1
        return timedelta(
            weeks=sign * int(m.group("weeks") or 0),
            days=sign * int(m.group("days") or 0),
            hours=sign * int(m.group("hours") or 0),
            minutes=sign * int(m.group("minutes") or 0),
            seconds=sign * int(m.group("seconds") or 0),
            milliseconds=sign * int(m.group("milliseconds") or 0),
        )


def readable_timedelta(
    *,
    strip_whitespace: bool = True,
    to_lower: bool = True,
    needs_sign: bool = False,
    allow_negative: bool = True,
    allow_dimensionless_seconds: bool = True,
) -> Type[timedelta]:
    namespace = {
        "strip_whitespace": strip_whitespace,
        "to_lower": to_lower,
        "needs_sign": needs_sign,
        "allow_negative": allow_negative,
        "allow_dimensionless_seconds": allow_dimensionless_seconds,
    }
    return _registered(type("TimeDeltaValue", (_TimeDelta,), namespace))


ConfigTimeDelta = readable_timedelta(allow_negative=False)
APITimeDelta = readable_timedelta(needs_sign=True, allow_dimensionless_seconds=False)



import asyncio
from copy import deepcopy

from sqlalchemy import MetaData, create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.orm import declarative_base, sessionmaker

from ..configuration import config
from ..exceptions import DuffyConfigurationError


# use custom metadata to specify naming convention
def constraint_column_names(constraint, table):
    return "_".join(c.name for c in constraint.columns)


def constraint_column_labels(constraint, table):  # pragma: no cover
    return "_".join(c._label for c in constraint.columns)


naming_convention = {
    "column_names": constraint_column_names,
    "column_labels": constraint_column_labels,
    "ix": "%(column_labels)s_index",
    "uq": "%(table_name)s_%(column_names)s_key",
    "ck": "%(table_name)s_%(constraint_name)s_check",
    "fk": "%(table_name)s_%(column_names)s_%(referred_table_name)s_fkey",
    "pk": "%(table_name)s_pkey",
}
metadata = MetaData(naming_convention=naming_convention)

Base = declarative_base(metadata=metadata)

async_session_maker = sessionmaker(class_=AsyncSession, expire_on_commit=False, future=True)
sync_session_maker = sessionmaker(future=True, expire_on_commit=False)


def init_sync_model(sync_engine: Engine = None):
    if not sync_engine:
        sync_engine = get_sync_engine()
    sync_session_maker.configure(bind=sync_engine)


async def init_async_model(async_engine: AsyncEngine = None):
    if not async_engine:
        async_engine = get_async_engine()
    async_session_maker.configure(bind=async_engine)


def init_model(sync_engine: Engine = None, async_engine: AsyncEngine = None):
    init_sync_model(sync_engine)
    asyncio.run(init_async_model(async_engine))


_key_failed_to_config_key = {
    "sqlalchemy": "database.sqlalchemy",
    "sync_url": "database.sqlalchemy.sync_url",
    "async_url": "database.sqlalchemy.async_url",
}


def get_sync_engine():
    try:
        sync_config = deepcopy(config["database"]["sqlalchemy"]) or {}
        sync_config["url"] = sync_config.pop("sync_url")
    except (AttributeError, KeyError) as exc:
        key_not_found = exc.args[0]
        raise DuffyConfigurationError(
            _key_failed_to_config_key.get(key_not_found, key_not_found)
        ) from exc
    sync_config.pop("async_url", None)
    return create_engine(**sync_config)


def get_async_engine():
    try:
        async_config = deepcopy(config["database"]["sqlalchemy"]) or {}
        async_config["url"] = async_config.pop("async_url")
    except (AttributeError, KeyError) as exc:
        key_not_found = exc.args[0]
        raise DuffyConfigurationError(
            _key_failed_to_config_key.get(key_not_found, key_not_found)
        ) from exc
    async_config.pop("sync_url", None)
    return create_async_engine(**async_config)



from typing import List

from sqlalchemy import JSON, Column, ForeignKey, Integer
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship

from ...api_models import SessionNodeModel
from .. import Base
from ..util import CreatableMixin, RetirableMixin, TZDateTime
from .tenant import Tenant


class Session(Base, CreatableMixin, RetirableMixin):
    __tablename__ = "sessions"
    __mapper_args__ = {"eager_defaults": True}
    id = Column(Integer, primary_key=True, nullable=False)
    tenant_id = Column(Integer, ForeignKey(Tenant.id), nullable=False)
    tenant = relationship(Tenant, backref="sessions")

    expires_at = Column(TZDateTime, nullable=True)

    data = Column(
        MutableDict.as_mutable(JSON), nullable=False, default=lambda: {}, server_default="{}"
    )

    session_nodes = relationship("SessionNode", back_populates="session")

    @property
    def nodes(self) -> List[SessionNodeModel]:
        """Combine info from related session_nodes and their nodes."""
        return [sn.pydantic_view for sn in self.session_nodes]



class DuffyException(Exception):  # pragma: no cover
    """Custom exceptions for Duffy."""


class DuffyConfigurationError(DuffyException):  # pragma: no cover
    """Something's wrong with the configuration of Duffy."""


class DuffyShellUnavailableError(DuffyException):  # pragma: no cover
    """The selected interactive shell type is not available."""



import sys
import uuid
from pathlib import Path

import alembic.command
import alembic.config
from sqlalchemy import inspect

from ..configuration import config

# Import the DB model here so its classes are considered by metadata.create_all() below.
from . import get_sync_engine, metadata, model, sync_session_maker  # noqa: F401

HERE = Path(__file__).parent


def setup_db_schema():
    engine = get_sync_engine()

    inspection_result = inspect(engine)

    present_tables = sorted(n for n in metadata.tables if inspection_result.has_table(n))

    if present_tables:
        print(f"Tables already present: {', '.join(present_tables)}", file=sys.stderr)
        print("Refusing to change database schema.", file=sys.stderr)
        sys.exit(1)

    with engine.begin():
        print("Creating database schema")
        metadata.create_all(bind=engine)

        print("Setting up database migrations")
        cfg = alembic.config.Config()
        cfg.set_main_option("script_location", str(HERE / "migrations"))
        cfg.set_main_option("sqlalchemy.url", config["database"]["sqlalchemy"]["sync_url"])

        alembic.command.stamp(cfg, "head")


def _gen_test_api_key(tenant_name):
    """Generate deterministic API keys for test users."""
    return uuid.uuid5(uuid.NAMESPACE_OID, tenant_name)


def _gen_test_data_objs():
    objs = set()

    admin = model.Tenant(
        name="admin", api_key=_gen_test_api_key("admin"), ssh_key="Boo", is_admin=True
    )
    objs.add(admin)

    tenant = model.Tenant(name="tenant", api_key=_gen_test_api_key("tenant"), ssh_key="Boo!")
    objs.add(tenant)

    node_specs = [
        {
            "hostname": "node-seamicro-1.example.net",
            "ipaddr": "192.168.0.11",
            "pool": "physical-centos8stream-x86_64",
            "data": {"architecture": "x86_64"},
        },
        {
            "hostname": "node-seamicro-2.example.net",
            "ipaddr": "192.168.0.12",
            "pool": "physical-fedora35-x86_64",
            "data": {"architecture": "x86_64"},
        },
        {
            "hostname": "node-seamicro-3.example.net",
            "ipaddr": "192.168.0.13",
            "data": {"architecture": "x86_64"},
        },
        {
            "hostname": "node-seamicro-4.example.net",
            "ipaddr": "192.168.0.14",
            "data": {"architecture": "x86_64"},
        },
    ]

    for quantity, flavour, ipaddrbase in (
        (40, "small", "192.168.1.10"),
        (20, "medium", "192.168.2.10"),
        (10, "large", "192.168.3.10"),
    ):
        ipprefix, lastoctetbase = ipaddrbase.rsplit(".", 1)
        lastoctetbase = int(lastoctetbase)
        for index in range(1, quantity + 1):
            lastoctet = lastoctetbase + index
            if index % 2:
                distro = "fedora35"
            else:
                distro = "centos8stream"
            node_specs.append(
                {
                    "hostname": f"node-opennebula-{flavour}-{index}.example.net",
                    "ipaddr": f"{ipprefix}.{lastoctet}",
                    "pool": f"virtual-{distro}-x86_64-{flavour}",
                }
            )

    for node_spec in node_specs:
        if node_spec.get("pool"):
            state = "ready"
        else:
            state = "unused"
        node = model.Node(state=state, **node_spec)
        objs.add(node)

    return objs


def setup_db_test_data():
    db_sync_session = sync_session_maker()
    print("Creating test data")
    with db_sync_session.begin():
        for obj in _gen_test_data_objs():
            db_sync_session.add(obj)

    print("Caution! Created tenants with deterministic API keys:")
    print(f"\tadmin:{_gen_test_api_key('admin')}")
    print(f"\ttenant:{_gen_test_api_key('tenant')}")
    print("Don't use in production!")



from typing import Any, Dict, Iterator, List, Optional, Union

import jinja2

from ..configuration import config
from ..database.model import Node
from ..util import merge_dicts
from .mechanisms import Mechanism


class NodePool(dict):

    subcls_per_cls_type: Dict[str, type] = {}
    known_pools: Dict[str, "NodePool"] = {}

    def __init_subclass__(cls, cls_type: bool, **kwargs):
        if cls_type in cls.subcls_per_cls_type:
            raise TypeError(f"Subclass type isn't unique: {cls_type}")

        cls.subcls_per_cls_type[cls_type] = cls
        cls.cls_type = cls_type

    def __init__(
        self, *, name: str, extends: Optional[Union[List[str], str]] = None, **configuration
    ):
        if name in self.known_pools:
            raise ValueError(f"Pool {name} is defined already")

        self.name = name

        if not extends:
            extends = []
            merged_configuration = configuration
        else:
            if isinstance(extends, str):
                extends = [extends]
            configs_to_be_merged = [self.known_pools[name] for name in extends] + [configuration]
            merged_configuration = merge_dicts(*configs_to_be_merged)

        self.extends = extends

        merged_configuration.pop("name", None)
        super().__init__(name=name, **merged_configuration)

        self.known_pools[name] = self

    def __str__(self):
        return str(dict(self))

    def __repr__(self):
        filtered_dict = {k: v for k, v in self.items() if k != "name"}
        return (
            f"{type(self).__name__}(name={self.name!r}, extends={self.extends!r},"
            + f" **{filtered_dict!r})"
        )

    @classmethod
    def process_configuration(cls):
        for cls_type, pool_configurations in config["nodepools"].items():
            for pool_name, pool_configuration in pool_configurations.items():
                pool_configuration = pool_configuration.copy()
                extends = pool_configuration.pop("extends", None)
                cls.subcls_per_cls_type[cls_type](
                    name=pool_name,
                    extends=extends,
                    **pool_configuration,
                )

    @classmethod
    def iter_pools(cls) -> Iterator["NodePool"]:
        for pool in cls.known_pools.values():
            yield pool

    def render_template(self, template: str, overrides: Optional[Dict[str, Any]] = None) -> str:
        template_vars = dict(self)
        if overrides:
            template_vars = {**self, **overrides}
        return jinja2.Template(template).render(**template_vars)

    def render_templates_in_obj(self, obj: Any, overrides: Optional[Dict[str, Any]] = None) -> Any:
        if isinstance(obj, str):
            return self.render_template(obj, overrides)
        elif isinstance(obj, dict):
            return {
                key: self.render_templates_in_obj(value, overrides) for key, value in obj.items()
            }
        else:
            return obj


class AbstractNodePool(NodePool, cls_type="abstract"):
    pass


class ConcreteNodePool(NodePool, cls_type="concrete"):
    def __init__(
        self, *, name: str, extends: Optional[Union[List[str], str]] = None, **configuration
    ):
        super().__init__(name=name, extends=extends, **configuration)

        self.mechanism = Mechanism.from_configuration(self, self["mechanism"])

    @classmethod
    def iter_pools(cls) -> Iterator["ConcreteNodePool"]:
        for pool in super().iter_pools():
            # filter for pools of this class or children
            if isinstance(pool, cls):
                yield pool

    def provision(self, nodes: List[Node]) -> Dict[str, Any]:
        return self.mechanism.provision(nodes=nodes)

    def deprovision(self, nodes: List[Node]) -> Dict[str, Any]:
        return self.mechanism.deprovision(nodes=nodes)



from functools import wraps

from pottery import Redlock
from redis import Redis

from ..configuration import config


class Lock(Redlock):
    """Redlock, using Duffy configuration."""

    @wraps(Redlock.__init__)
    def __init__(self, *, masters=None, **kwargs):
        if not masters:
            masters = {Redis.from_url(config["tasks"]["locking"]["url"])}
        super().__init__(masters=masters, **kwargs)



from typing import TYPE_CHECKING, Any, Dict, List

from ...database.model import Node

if TYPE_CHECKING:
    from ..node_pools import NodePool


class MechanismFailure(Exception):
    pass


class Mechanism(dict):

    known_mechanisms: Dict[str, "Mechanism"] = {}

    def __init_subclass__(cls, mech_type: str, **kwargs):
        super().__init_subclass__()

        # Register subclasses.
        if mech_type in cls.known_mechanisms:
            raise TypeError(f"Mechanism type isn't unique: {mech_type}")

        cls.known_mechanisms[mech_type] = cls
        cls.mech_type = mech_type

    def __init__(self, *, nodepool: "NodePool", **kwargs):
        super().__init__(**kwargs)
        self.nodepool = nodepool

    @classmethod
    def from_configuration(cls, nodepool: "NodePool", configuration: Dict[str, Any]):
        configuration = configuration.copy()
        mech_type = configuration.pop("type")
        return cls.known_mechanisms[mech_type](nodepool=nodepool, **configuration[mech_type])

    def provision(self, nodes: List[Node]) -> Dict[str, Any]:
        raise NotImplementedError()

    def deprovision(self, nodes: List[Node]) -> Dict[str, Any]:
        raise NotImplementedError()



import re
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import AnyUrl, BaseModel, Field, RedisDsn, conint, stricturl, validator

from ..misc import ConfigTimeDelta

# enums


class LogLevel(str, Enum):
    trace = "trace"
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"


class MechanismType(str, Enum):
    ansible = "ansible"


# Pydantic models


class ConfigBaseModel(BaseModel):
    class Config:
        extra = "forbid"


class CeleryModel(BaseModel):
    # This is intentionally not a subclass of ConfigBaseModel, it is passed on to Celery, i.e. can
    # contain arbitrarily named fields.
    broker_url: AnyUrl
    result_backend: AnyUrl


class LockingModel(ConfigBaseModel):
    url: RedisDsn


class PeriodicTaskModel(ConfigBaseModel):
    interval: ConfigTimeDelta


class TasksModel(ConfigBaseModel):
    celery: CeleryModel
    locking: LockingModel
    periodic: Optional[Dict[str, PeriodicTaskModel]]


class SQLAlchemyModel(ConfigBaseModel):
    sync_url: stricturl(tld_required=False, host_required=False)
    async_url: stricturl(tld_required=False, host_required=False)


class DatabaseModel(ConfigBaseModel):
    sqlalchemy: SQLAlchemyModel


class DefaultsModel(ConfigBaseModel):
    session_lifetime: ConfigTimeDelta = Field(alias="session-lifetime")
    session_lifetime_max: ConfigTimeDelta = Field(alias="session-lifetime-max")
    node_quota: conint(gt=0) = Field(alias="node-quota")


class AnsibleMechanismPlaybookModel(ConfigBaseModel):
    extra_vars: Optional[Dict[str, Any]] = Field(alias="extra-vars")
    playbook: Path


class AnsibleMechanismModel(ConfigBaseModel):
    topdir: Optional[Path]
    extra_vars: Optional[Dict[str, Any]] = Field(alias="extra-vars")
    provision: Optional[AnsibleMechanismPlaybookModel]
    deprovision: Optional[AnsibleMechanismPlaybookModel]


class MechanismModel(ConfigBaseModel):
    type_: Optional[MechanismType] = Field(alias="type")
    ansible: Optional[AnsibleMechanismModel]


class NodePoolsModel(ConfigBaseModel):
    extends: Optional[str]
    mechanism: Optional[MechanismModel]
    fill_level: Optional[conint(gt=0)] = Field(alias="fill-level")
    reuse_nodes: Optional[Union[Dict[str, Union[int, str]], Literal[False]]] = Field(
        alias="reuse-nodes"
    )
    run_parallel: Optional[bool] = Field(alias="run-parallel")

    class Config:
        extra = "allow"


class NodePoolsRootModel(ConfigBaseModel):
    abstract: Optional[Dict[str, NodePoolsModel]]
    concrete: Dict[str, NodePoolsModel]


class LoggingModel(ConfigBaseModel):
    version: Literal[1]

    class Config:
        extra = "allow"


class AppModel(ConfigBaseModel):
    loglevel: Optional[LogLevel]
    host: Optional[str]
    port: Optional[conint(gt=0, lt=65536)]
    logging: Optional[LoggingModel]


class LegacyPoolMapModel(ConfigBaseModel):
    pool: str
    ver: Optional[str]
    arch: Optional[str]
    flavor: Optional[str]

    @validator("ver", "arch", "flavor")
    def detect_regex(cls, v: str):
        if v.startswith("^") and v.endswith("$"):
            return re.compile(v)
        return v


class LegacyModel(ConfigBaseModel):
    host: Optional[str]
    port: Optional[conint(gt=0, lt=65536)]
    dest: Optional[str]
    loglevel: Optional[LogLevel]
    logging: Optional[LoggingModel]
    usermap: Dict[str, str]
    poolmap: List[LegacyPoolMapModel]


class ConfigModel(ConfigBaseModel):
    app: Optional[AppModel]
    tasks: Optional[TasksModel]
    database: Optional[DatabaseModel]
    defaults: Optional[DefaultsModel]
    metaclient: Optional[LegacyModel]
    nodepools: Optional[NodePoolsRootModel]



from copy import deepcopy
from itertools import chain
from pathlib import Path
from typing import List, Union

import yaml

from ..util import merge_dicts
from .validation import ConfigModel

config = {}


def _expand_normalize_config_files(config_files: List[Union[Path, str]]) -> List[Path]:
    config_file_paths = []

    for path in config_files:
        if not isinstance(path, Path):
            path = Path(path)
        if path.is_dir():
            config_file_paths.extend(sorted(chain(path.glob("*.yaml"), path.glob("*.yml"))))
        else:
            config_file_paths.append(path)

    return config_file_paths


def read_configuration(
    *config_files: List[Union[Path, str]], clear: bool = True, validate: bool = True
):
    config_files = _expand_normalize_config_files(config_files)

    if clear:
        new_config = {}
    else:
        new_config = deepcopy(config)

    for config_file in config_files:
        with config_file.open("r") as fp:
            for config_doc in yaml.safe_load_all(fp):
                new_config = merge_dicts(new_config, config_doc)

    if validate:
        # validate merged configuration
        ConfigModel(**new_config)

    config.clear()
    config.update(new_config)



import uuid
from functools import lru_cache

import bcrypt
from sqlalchemy import Boolean, Column, Integer, Interval, Text, UnicodeText, text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.sql import case

from ...configuration import config
from ...configuration.validation import DefaultsModel
from .. import Base
from ..util import CreatableMixin, RetirableMixin


@lru_cache
def _defaults_config():
    return DefaultsModel(**config["defaults"])


class Tenant(Base, CreatableMixin, RetirableMixin):
    __tablename__ = "tenants"
    __mapper_args__ = {"eager_defaults": True}
    id = Column(Integer, primary_key=True, nullable=False)
    name = Column(UnicodeText, unique=True, nullable=False)
    is_admin = Column(Boolean, nullable=False, default=False, server_default=text("FALSE"))
    ssh_key = Column(UnicodeText, nullable=False)
    _api_key = Column("api_key", Text, nullable=False)
    node_quota = Column(Integer, nullable=True)
    session_lifetime = Column(Interval, nullable=True)
    session_lifetime_max = Column(Interval, nullable=True)

    @hybrid_property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, key: uuid.UUID):
        salt = bcrypt.gensalt()
        hashed_key = bcrypt.hashpw(str(key).encode("ascii"), salt)
        self._api_key = hashed_key.decode("ascii")

    def validate_api_key(self, key: uuid.UUID):
        return bcrypt.checkpw(str(key).encode("ascii"), self._api_key.encode("ascii"))

    @hybrid_property
    def effective_node_quota(self):
        if self.node_quota is not None:
            return self.node_quota

        return config["defaults"]["node-quota"]

    @effective_node_quota.expression
    def effective_node_quota(cls):
        return case(
            [(cls.node_quota != None, cls.node_quota)],  # noqa: E711
            else_=_defaults_config().node_quota,
        )

    @hybrid_property
    def effective_session_lifetime(self):
        if self.session_lifetime is not None:
            return self.session_lifetime

        return _defaults_config().session_lifetime

    @effective_session_lifetime.expression
    def effective_session_lifetime(cls):
        return case(
            [(cls.session_lifetime != None, cls.session_lifetime)],  # noqa: E711
            else_=_defaults_config().session_lifetime,
        )

    @hybrid_property
    def effective_session_lifetime_max(self):
        if self.session_lifetime_max is not None:
            return self.session_lifetime_max

        return _defaults_config().session_lifetime_max

    @effective_session_lifetime_max.expression
    def effective_session_lifetime_max(cls):
        return case(
            [(cls.session_lifetime_max != None, cls.session_lifetime_max)],  # noqa: E711
            else_=_defaults_config().session_lifetime_max,
        )



import asyncio
from collections import defaultdict
from typing import List

from celery.utils.log import get_task_logger
from sqlalchemy import select

from ..database import sync_session_maker
from ..database.model import Node
from ..database.types import NodeState
from ..nodes_context import decontextualize
from .base import celery
from .mechanisms import MechanismFailure
from .node_pools import ConcreteNodePool, NodePool
from .provision import fill_pools

log = get_task_logger(__name__)

UNSET = object()

# These fields should be cleaned out when deprovisioning a reusable node.
NODE_DATA_EPHEMERAL_FIELDS = ("error", "nodes_spec", "provision")


@celery.task(bind=True)
def deprovision_pool_nodes(self, pool_name: str, node_ids: List[int]):
    log.debug("[%s] Deprovisioning nodes from pool (begin): %r", pool_name, node_ids)

    try:
        pool = NodePool.known_pools[pool_name]
    except KeyError:
        log.error("[%s] Can't find pool.", pool_name)
        return

    if not isinstance(pool, ConcreteNodePool):
        log.error("[%s] Pool must be a concrete node pool.", pool_name)
        return

    try:
        with sync_session_maker() as db_sync_session, db_sync_session.begin():
            nodes = (
                db_sync_session.execute(
                    select(Node)
                    .filter_by(active=True, pool=pool_name, state=NodeState.deployed)
                    .filter(Node.id.in_(node_ids))
                )
                .scalars()
                .all()
            )

            for node in nodes:
                node.state = NodeState.deprovisioning
                node.pool = None

        log.debug("Decontextualizing nodes.")
        # ignore results, after use anything could be broken on the nodes
        asyncio.run(decontextualize([node.ipaddr for node in nodes]))

        found_node_ids = {node.id for node in nodes}
        not_found_node_ids = set(node_ids) - found_node_ids

        if not_found_node_ids:
            log.warning(
                "[%s] Didn't find deployed nodes with ids: %s", pool_name, not_found_node_ids
            )

        log.debug("[%s] Attempting to deprovision nodes: %r", pool_name, found_node_ids)

        with sync_session_maker() as db_sync_session, db_sync_session.begin():
            nodes = [db_sync_session.merge(node, load=False) for node in nodes]

            try:
                deprov_result = pool.deprovision(nodes)
            except MechanismFailure:
                log.error("[%s] Deprovisioning mechanism failed.", pool.name)
                log.debug("[%s] Marking nodes as failed in database.", pool.name)
                with sync_session_maker() as db_sync_session_in_exc, db_sync_session_in_exc.begin():
                    for node in nodes:
                        exc_node = db_sync_session_in_exc.merge(node, load=False)
                        exc_node.fail("deprovisioning mechanism failed")
                raise

            unmatched_nodes = set(nodes)
            matched_nodes = set()

            for node_res in deprov_result["nodes"]:
                # match up nodes with the data blurb from their provisioning
                for node in unmatched_nodes:
                    matched = False

                    for nr_key, nr_value in node_res.items():
                        if node.data.get("provision", {}).get(nr_key, UNSET) != nr_value:
                            matched = False
                            break

                        matched = True

                    if matched:
                        # at least one field considered was matched
                        matched_nodes.add(node)
                        unmatched_nodes.remove(node)
                        break
                else:
                    # didn't break out of loop -> no node object matched
                    log.warning("[%s] Node result couldn't be matched: %r", pool.name, node_res)

            if unmatched_nodes:
                # handle & report nodes which apparently weren't deprovisioned
                unmatched_ids = []
                for node in unmatched_nodes:
                    unmatched_ids.append(node.id)
                    node.fail("deprovisioning node failed")

                log.warning("[%s] Nodes unmatched in result: %r", pool.name, sorted(unmatched_ids))

            # clean up DB objects of deprovisioned nodes
            for node in matched_nodes:
                if node.reusable:
                    node.state = NodeState.unused
                    for fname in NODE_DATA_EPHEMERAL_FIELDS:
                        node.data.pop(fname, None)
                else:
                    node.state = NodeState.done
                    node.active = False

            if any(node.reusable for node in matched_nodes):
                fill_pools.delay().forget()

    except Exception:
        log.error("[%s] Deprovisioning failed: %r", pool_name, node_ids)
        raise

    log.debug("[%s] Deprovisioning nodes from pool (end): %r", pool_name, node_ids)


@celery.task
def deprovision_nodes(node_ids: List[int]):
    """Deprovision nodes e.g. of an expired session.

    This divides up nodes by their pools and kicks off sub tasks for
    each pool or each node, depending on the respective `run-parallel` setting
    of the pool.
    """
    log.debug("deprovision_nodes(%r) begin", node_ids)
    pools_node_ids = defaultdict(list)

    with sync_session_maker() as db_sync_session, db_sync_session.begin():
        # First, find the -- active, deployed -- nodes with the supplied ids, sort them by their
        # pools and change their state to 'deprovisioning', then kick off sub tasks which
        # deprovision all nodes that belong to the same pool.
        nodes = (
            db_sync_session.execute(
                select(Node)
                .filter_by(active=True, state=NodeState.deployed)
                .filter(Node.id.in_(node_ids))
            )
            .scalars()
            .all()
        )

        found_node_ids = {node.id for node in nodes}
        not_found_node_ids = sorted(set(node_ids) - found_node_ids)

        if not_found_node_ids:
            log.warning("Didn't find deployed nodes with ids: %s", not_found_node_ids)

        for node in nodes:
            try:
                pool = NodePool.known_pools[node.pool]
            except KeyError:
                log.error("[%s] Pool not found for node with id: %d", node.pool, node.id)
                node.fail(f"deprovisioning node failed, pool '{node.pool}' not found")
                continue

            pools_node_ids[node.pool].append(node.id)

    for pool_name, node_ids in pools_node_ids.items():
        pool = NodePool.known_pools[pool_name]
        log.debug("Creating task(s) to deprovision session nodes in pool %s", pool.name)
        if pool.get("run-parallel", True):
            for node_id in node_ids:
                deprovision_pool_nodes.delay(pool_name=pool.name, node_ids=[node_id]).forget()
        else:
            deprovision_pool_nodes.delay(pool_name=pool.name, node_ids=node_ids).forget()

    log.debug("deprovision_nodes(%r) end", node_ids)
